/*	$Id: ImageDecoderFactory.cpp,v 1.1 2001/10/07 18:23:11 kent Exp $	-*- c++ -*- */
/*
 * Copyright 2001 TAMURA Kent
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <new.h>
#include "ImageDecoderFactory.h"
#include "Rip.h"
#include "Rqd.h"

bool
ImageDecoderFactory::isSupported(const char* filename, const uint8_t* buffer, size_t size)
{
        if (Rip::isSupported(filename, buffer, size))
                return true;
        if (Rqd::isSupported(filename, buffer, size))
                return true;
        return false;
}


ImageDecoder*
ImageDecoderFactory::create(const uint8_t* buffer, size_t size)
{
        ImageDecoder* decoder = NULL;
        try {
                if (Rip::isSupported("", buffer, size)) {
                        decoder = new Rip;
                } else if (Rqd::isSupported("", buffer, size)) {
                        decoder = new Rqd;
                }
                if (decoder && !decoder->initialize(buffer, size)) {
                        delete decoder;
                        return NULL;
                }
        } catch (bad_alloc) {
        }
        return decoder;
}

