/*	$Id: RioArchive.h,v 1.5 2001/10/07 15:39:18 kent Exp $	-*- c++ -*- */
/*
 * Copyright 2001 TAMURA Kent
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef __RIOARCHIVE_H__
#define __RIOARCHIVE_H__

#include <stdio.h>
#include "config.h"
#include "NameFilter.h"
#include "Stream.h"
#ifdef _WIN32
# include "susie.h"
#endif


class ClassMap;
class ClassDescriptor;
class ObjectOceanNode;
class RioArchive;


class RioObject {
protected:
	const ClassDescriptor* desc;

public:
        RioObject(const ClassDescriptor* desc);
        virtual ~RioObject();
        virtual void dump() const = 0;
        virtual void setParent(ObjectOceanNode* parent) = 0;
};


class ObjectOcean: public RioObject {
	uint16_t numObjects;
	ObjectOceanNode **nodes;

public:
        ObjectOcean(const ClassDescriptor* d, RioArchive* rio, uint16_t num);
        virtual ~ObjectOcean();
        virtual void dump() const;
        virtual void setParent(ObjectOceanNode* parent);
                                // for Susie
        int getTotalNum(const NameFilter* filter) const;
        const ObjectOceanNode* searchFor(const char* name, bool ignoreCase) const;
        const ObjectOceanNode* searchFor(uint32_t offset) const;
#ifdef _WIN32
        SusieFileInfo* storeAll(const NameFilter* filter, SusieFileInfo* sfi) const;
#endif
};


class RioArchive {
	Stream* stream;
	ClassMap *top;
	uint32_t currentCount;
        RioObject* topObject;
        const ClassDescriptor* cdesc[3];
        const ClassMap* nullClassMap;

protected:
        const ClassDescriptor* searchStaticsFor(const char* cname);

public:
        /**
         * œnꂽ Stream IuWFNǵA RioArchive IuWFNg
         * jBāAŃX^bN Stream IuWFNgnĂ͂
         * ȂB
         */
        RioArchive(Stream* st);
        ~RioArchive();
        RioObject* load();
        const ClassMap* loadClass();
        RioObject* getTopObject() const {
                return this->topObject;
        }
        void extract(const ObjectOceanNode* node, void* buf);
        uint8_t get8() {
                return this->stream->get8();
        }
        uint16_t get16le() {
                return this->stream->get16le();
        }
        uint32_t get32le() {
                return this->stream->get32le();
        }
        void seek(int off) {
                this->stream->seek(off, SEEK_CUR);
        }
        int read(uint8_t* buf, size_t len) {
                return this->stream->read(buf, len);
        }
};


class ObjectOceanNode {
        RioObject* robj;
        const ClassDescriptor* desc; // backup of robj->desc
	char *name;
	uint32_t begin;
	uint32_t size;
        ObjectOceanNode* parent;

public:
        ObjectOceanNode(RioArchive* rio);
        ~ObjectOceanNode();
        void dump();
        bool hasChildren() const {
                return this->robj != NULL;
        }
        RioObject* getObject() const {
                return this->robj;
        }
        const char* getName() const {
                return this->name;
        }

        uint32_t getOffset() const {
                return this->begin;
        }
        uint32_t getSize() const {
                return this->size;
        }
        void setParent(ObjectOceanNode* parent) {
                this->parent = parent;
        }

        int getTotalNum(const NameFilter* filter) const;
        int storePath(char* path, const int len) const;
#ifdef _WIN32
        SusieFileInfo* store(const NameFilter* filter, SusieFileInfo* sfi) const;
#endif
};

#endif // __RIOARCHIVE_H__
