/*	$Id: Rip.cpp,v 1.16 2001/12/15 14:19:12 kent Exp $	-*- c++ -*- */
/*
 * Copyright (c) 2001 Takuya SHIOZAKI <AoiMoe@imou.to>, TAMURA Kent
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include "Rip.h"
#include "StreamImpl.h"

#undef USE_STATICTABLE
#define USE_CLIPPED             // Uses minimum bound rectangle for output DIB

/* ================================================================
 * for CRip007
 * ================================================================ */

class BitStream {
        const uint8_t* istream;
        uint32_t mask;
public:
        BitStream(const uint8_t* i) {
                this->istream = i;
                this->mask = 0x100;
        }

        inline bool get() {
                this->mask >>= 1;
                if (!this->mask) {
                        this->istream++;
                        this->mask = 0x80;
                }
                return !!(*this->istream & this->mask);
        }

        inline uint32_t get(int len) {
                uint32_t v = 0;
                while (len-- > 0) {
                        v <<= 1;
                        this->mask >>= 1;
                        if (!this->mask) {
                                this->istream++;
                                this->mask = 0x80;
                        }
                        v |= (*this->istream & this->mask) ? 1 : 0;
                }
                return v;
        }

        inline int32_t getSigned() {
                int32_t val = 0;
                bool saved;

                if (this->get()) {
                        val = 1;
                        saved = this->get();
                        while (this->get()) {
                                val <<= 1;
                                if (this->get())
                                        val |= 1;
                        }
                        if (saved)
                                val = -val;
                }
                return val;
        }

        inline uint32_t getUnsigned() {
                uint32_t val = 1;

                while (this->get()) {
                        val <<= 1;
                        if (this->get())
                                val |= 1;
                }
                return val;
        }
};

typedef uint32_t RipPixel;
#ifdef USE_STATICTABLE
# include "table.c"
#else
static bool isTableInitialized = false;
static uint8_t colorTranslationTable[6][256];
#endif
#define RIP_AMASK 0xFF000000
#define RIP_RMASK 0x00FF0000
#define RIP_GMASK 0x0000FF00
#define RIP_BMASK 0x000000FF
#define RIP_ASHIFT 24
#define RIP_RSHIFT 16
#define RIP_GSHIFT 8
#define RIP_BSHIFT 0
#define RIP_IBITS 8
#define RIP_RGBA1_AMAX 0x1F
#define RIP_RGBA1_ASHIFT 0x1B
#define RIP_GET_(c_,p_,s_) (((p_)&RIP_##c_##MASK)>>s_)
#define RIP_GET8_(c_,p_) RIP_GET_(c_,p_,RIP_##c_##SHIFT)
#define RIP_GETA8(p_) RIP_GET8_(A,p_)
#define RIP_GETR8(p_) RIP_GET8_(R,p_)
#define RIP_GETG8(p_) RIP_GET8_(G,p_)
#define RIP_GETB8(p_) RIP_GET8_(B,p_)
#define RIP_GETN_(c_,p_) RIP_GET_(c_,p_,shift##c_)
#define RIP_GETA(p_) RIP_GETN_(A,p_)
#define RIP_GETR(p_) RIP_GETN_(R,p_)
#define RIP_GETG(p_) RIP_GETN_(G,p_)
#define RIP_GETB(p_) RIP_GETN_(B,p_)
#define RIP_MAKE_S__(c_,p_,s_) ((p_)<<s_)
#define RIP_MAKE_S_8_(c_,p_) RIP_MAKE_S__(c_,p_,RIP_##c_##SHIFT)
#define RIP_MAKE_S_A8(p_) RIP_MAKE_S_8_(A,p_)
#define RIP_MAKE_S_R8(p_) RIP_MAKE_S_8_(R,p_)
#define RIP_MAKE_S_G8(p_) RIP_MAKE_S_8_(G,p_)
#define RIP_MAKE_S_B8(p_) RIP_MAKE_S_8_(B,p_)
#define RIP_MAKE_S_N_(c_,p_) RIP_MAKE_S__(c_,p_,shift##c_)
#define RIP_MAKE_S_A(p_) RIP_MAKE_S_N_(A,p_)
#define RIP_MAKE_S_R(p_) RIP_MAKE_S_N_(R,p_)
#define RIP_MAKE_S_G(p_) RIP_MAKE_S_N_(G,p_)
#define RIP_MAKE_S_B(p_) RIP_MAKE_S_N_(B,p_)
#define RIP_TABLECONV(v_) (((v_)>=0)?table[(v_)&0xFF]:-table[-(v_)&0xFF])

static inline int32_t
rip_clip(int32_t v, int32_t low, int32_t high)
{
	if (v<low)
		v = low;
	else if (v>high)
		v = high;
	return v;
}

void
Rip::decode007(RipPixel *dst, size_t bytes_per_line,
               int dstx, int dsty) const
{
	RipPixel *dstp;
	RipPixel defaultPixel, pixel;
	int shiftR = RIP_RSHIFT + RIP_IBITS - this->weightR;
	int shiftG = RIP_GSHIFT + RIP_IBITS - this->weightG;
	int shiftB = RIP_BSHIFT + RIP_IBITS - this->weightB;
	int x, y;
	int count;
	int32_t deltaG, deltaB, deltaR;
	bool isHalfDeltaG;
	int32_t tmp1, tmp2, tmpG;
	const uint8_t *table;
	BitStream bs(this->in);
        bool bottomup = true;

	defaultPixel =
	    RIP_AMASK |
	    ((0xFF >> this->weightR) << RIP_RSHIFT) |
	    ((0xFF >> this->weightG) << RIP_GSHIFT) |
	    ((0xFF >> this->weightB) << RIP_BSHIFT);
	table = colorTranslationTable[this->compInfo[CI_TABLE]];
	isHalfDeltaG = this->weightB == 6 && this->weightG == 7 && this->weightR == 6;

	dst += dstx;
        if (bottomup)
                dst += (this->getHeight() - 1) * (bytes_per_line/sizeof(RipPixel));
        else
                dst += dsty * (bytes_per_line/sizeof(RipPixel));

	for (y = 0; y < this->clipheight; y++) {
		pixel = defaultPixel;
		dstp = dst;
		for (x = 0; x < this->clipwidth;) {
			count = bs.getUnsigned();
			x+=count;
			for (; count>0; count--) {
				if (bs.get()) {
                                        if (bottomup)
                                                pixel = *(dstp
                                                          + bytes_per_line/sizeof(RipPixel));
                                        else
                                                pixel = *(dstp
                                                          - bytes_per_line/sizeof(RipPixel));
				} else {
					deltaG = bs.getSigned();
					deltaB = bs.getSigned();
					deltaB = RIP_TABLECONV(deltaB);
					deltaR = bs.getSigned();
					deltaR = RIP_TABLECONV(deltaR);
					tmpG = deltaG;
					if (isHalfDeltaG)
						tmpG >>= 1;
                                        
					if (this->compInfo[CI_VALCLIP]) {
						tmp1 =-RIP_GETB(pixel);
						tmp2 =RIP_GETB(RIP_BMASK)+tmp1;
						deltaB +=
						    rip_clip(tmpG, tmp1, tmp2);

						tmp1 =-RIP_GETR(pixel);
						tmp2 =RIP_GETR(RIP_RMASK)+tmp1;
						deltaR +=
						    rip_clip(tmpG, tmp1, tmp2);
					} else {
						deltaB += tmpG;
						deltaR += tmpG;
					}
					pixel += RIP_MAKE_S_R(deltaR);
					pixel += RIP_MAKE_S_G(deltaG);
					pixel += RIP_MAKE_S_B(deltaB);
				}
				*dstp++ = pixel;
			}
			if (x >= this->clipwidth) {
				break;
			}
			count = bs.getUnsigned();
			x+=count;
			for (; count>0; count--) {
				*dstp++ = pixel;
			}
		}
                if (bottomup)
                        dst -= bytes_per_line/sizeof(RipPixel);
                else
                        dst += bytes_per_line/sizeof(RipPixel);
	}
}


void
Rip::decode007A(RipPixel *dst, size_t bytes_per_line,
                int dstx, int dsty) const
{
	RipPixel *dstptr;
	RipPixel defaultPixel, pixel;
	RipPixel *tmpbuf;
	int shiftR = RIP_RSHIFT + RIP_IBITS - this->weightR;
	int shiftG = RIP_GSHIFT + RIP_IBITS - this->weightG;
	int shiftB = RIP_BSHIFT + RIP_IBITS - this->weightB;
	int x, y;
	bool isrun;
	int countA, runlength;
	int32_t alpha, deltaA, deltaR, deltaG, deltaB, tmp1, tmp2;
	const uint8_t *table;
	BitStream bs(this->in);
        bool bottomup = true;

	defaultPixel =
	    ((0xFF >> this->weightR) << RIP_RSHIFT) |
	    ((0xFF >> this->weightG) << RIP_GSHIFT) |
	    ((0xFF >> this->weightB) << RIP_BSHIFT);
	table = colorTranslationTable[this->compInfo[CI_TABLE]];
	tmpbuf = (RipPixel*)calloc(this->clipwidth, sizeof(RipPixel));
        if (tmpbuf == NULL)
                return;
        if (bottomup)
                dst += dstx + (this->getHeight() - 1)*(bytes_per_line/sizeof(RipPixel));
        else
                dst += dstx + dsty*(bytes_per_line/sizeof(RipPixel));

	for (y = 0; y < this->clipheight; y++) {
		dstptr = dst;
		countA = 0;
		pixel = 0;
		alpha = 0;
		isrun = true;
		runlength = 0;
		for (x = 0; x < this->clipwidth; ) {
			if (countA == 0) {
				deltaA = bs.getSigned();
				alpha += deltaA;
				if (alpha==0 || alpha==RIP_RGBA1_AMAX) {
					countA = bs.getUnsigned();
				}
			}
			if (alpha==0) {
				x+=countA;
				dstptr+=countA;
				countA=0;
				continue;
			}
			if (alpha==RIP_RGBA1_AMAX)
				countA--;
			if (runlength==0) {
				runlength = bs.getUnsigned();
				isrun = !isrun;
			}
			runlength--;
			if (!isrun) {
				if (bs.get()) {
					pixel = tmpbuf[x];
				} else {
					deltaG = bs.getSigned();
					deltaB = bs.getSigned();
					deltaB = RIP_TABLECONV(deltaB);
					deltaR = bs.getSigned();
					deltaR = RIP_TABLECONV(deltaR);

					tmp1 = -RIP_GETB(pixel);
					tmp2 = RIP_GETB(RIP_BMASK)+tmp1;
					deltaB += rip_clip(deltaG, tmp1, tmp2);

					tmp1 = -RIP_GETR(pixel);
					tmp2 = RIP_GETR(RIP_RMASK)+tmp1;
					deltaR += rip_clip(deltaG, tmp1, tmp2);

					pixel += RIP_MAKE_S_R(deltaR);
					pixel += RIP_MAKE_S_G(deltaG);
					pixel += RIP_MAKE_S_B(deltaB);
				}
			}
			tmp1 = defaultPixel+pixel;
			if (alpha==RIP_RGBA1_AMAX) {
				tmp1 |= RIP_AMASK;
			} else {
				tmp1 |= alpha << RIP_RGBA1_ASHIFT;
			}
			*dstptr++ = tmp1;
			tmpbuf[x++] = pixel;
		}
                if (bottomup)
                        dst -= bytes_per_line/sizeof(RipPixel);
                else
                        dst += bytes_per_line/sizeof(RipPixel);
	}
	free(tmpbuf);
}


/* ================================================================
 * for CRip
 * ================================================================ */

static int
decode6(uint32_t src, BitStream* stream)
{
        src &= 0xfc;
        switch (src) {
        case 0:
                if (!stream->get())
                        return src;
                if (!stream->get())
                        return stream->get(6) << 2;
                if (!stream->get())
                        return 4;
                return stream->get() ? 0x0c : 8;

        case 4:
                if (!stream->get())
                        return stream->get() ? 4 : 0;
                if (!stream->get())
                        return stream->get(6) << 2;
                if (!stream->get())
                        return 8;
                return stream->get() ? 0x10 : 0x0c;

        case 0xfc:
                if (!stream->get())
                        return 0xfc;
                if (!stream->get())
                        return stream->get(6) << 2;
                if (!stream->get())
                        return 0xf8;
                return stream->get() ? 0xf0 : 0xf4;
        }

        if (!stream->get())
                return stream->get() ? (stream->get(6) << 2) : src;
        if (!stream->get())
                return stream->get() ? src - 4 : src + 4;
        if (!stream->get())
                return stream->get() ? src - 8 : src + 8;

        switch (stream->get(2)) {
        case 0:
                src += 0x10;
                return src > 0xfc ? 0xfc : src;
        case 1:
                return src < 0x10 ? 0 : src - 0x10;
        case 2:
                src += 0x18;
                return src > 0xfc ? 0xfc : src;
        default:                // Actually, candidate value is only `3'.
                return src < 0x18 ? 0 : src - 0x18;
        }
}


static int
decode7(uint32_t src, BitStream* stream)
{
        src &= 0xfe;
        if (!stream->get())
                return stream->get() ? (stream->get(6) << 2) : src;
        if (!stream->get())
                return stream->get() ? src - 2 : src + 2;
        if (!stream->get())
                return stream->get() ? src - 4 : src + 4;

        switch (stream->get(2)) {
        case 0:
                src += 8;
                return src > 0xfe ? 0xfe : src;
        case 1:
                return src < 8 ? 0 : src - 8;
        case 2:
                src += 0x0c;
                return src > 0xfe ? 0xfe : src;
        default:
                return src < 0x0c ? 0 : src - 0x0c;
        }
}

static inline RipPixel
updatePixelRgb3(BitStream* stream, RipPixel pixel)
{
        RipPixel newpixel;
        uint32_t elem;

        if (stream->get()) {
                newpixel = decode6(pixel, stream) + 3;
                elem = decode7(pixel >> RIP_GSHIFT, stream) + 1;
                newpixel |= elem << RIP_GSHIFT;
                elem = decode6(pixel >> RIP_RSHIFT, stream) + 3;
                newpixel |= elem << RIP_RSHIFT;
                pixel = newpixel;
        }
        return pixel;
}


void
Rip::decodeRgb3(uint32_t* dst, size_t bytes_per_line) const
{
        RipPixel pixel = 0;
        bool bottomup = true;
        BitStream stream(this->in);

        for (int y = this->clipheight-1; y >= 0; y--) {
                uint32_t* writer;

                if (bottomup)
                        writer = dst + (this->clipheight - y - 1)
                                * bytes_per_line/sizeof(uint32_t);
                else
                        writer = dst + y * bytes_per_line/sizeof(uint32_t);
                for (int x = 0; x < this->clipwidth; x++) {
                        pixel = updatePixelRgb3(&stream, pixel);
                        *writer++ = pixel | RIP_AMASK;
                }
        }
}


static int
decode6a(uint32_t src, BitStream* stream)
{
        src &= 0xfc;
        if (!stream->get())
                return stream->get() ? (stream->get(6) << 2) : src;
        if (!stream->get())
                return stream->get() ? src - 4 : src + 4;
        if (!stream->get())
                return stream->get() ? src - 8 : src + 8;

        switch (stream->get(2)) {
        case 0:
                src += 0x10;
                return src > 0xfc ? 0xfc : src;
        case 1:
                return src < 0x10 ? 0 : src - 0x10;
        case 2:
                src += 0x18;
                return src > 0xfc ? 0xfc : src;
        case 3:
        default:
                return src < 0x18 ? 0 : src - 0x18;
        }
}


static inline RipPixel
updatePixelRgba(BitStream* stream, RipPixel pixel)
{
        RipPixel newpixel;
        uint32_t elem;

        if (stream->get()) {
                newpixel = decode6a(pixel, stream) + 3;
                elem = decode6a(pixel >> RIP_GSHIFT, stream) + 3;
                newpixel |= elem << RIP_GSHIFT;
                elem = decode6a(pixel >> RIP_RSHIFT, stream) + 3;
                newpixel |= elem << RIP_RSHIFT;
                pixel = newpixel;
        }
        return pixel;
}

void
Rip::decodeRgba(uint32_t* dst, size_t bytes_per_line) const
{
        bool bottomup = true;
        BitStream stream(this->in + sizeof(uint32_t));
        const uint8_t* areader = this->in + *(uint32_t*)this->in;

        for (int y = 0; y < this->clipheight; y++) {
                int x = 0;
                uint32_t* writer;
                RipPixel pixel = 0;
                uint32_t alpha = 0;

                if (bottomup)
                        writer = dst + (this->clipheight - y -1)
                                * bytes_per_line/sizeof(uint32_t) + x;
                else
                        writer = dst + y * bytes_per_line/sizeof(uint32_t) + x;
                do {
                        uint32_t adata;

                        if (areader >= this->in + this->size) {
                                return;
                        }
                        adata = *areader++;
                        x += adata;
                        if (alpha == 0) { // alpha ̓[v̍ŏł0
                                writer += adata; // n sNZړ
                        } else {
                                while (adata-- > 0) {
                                        pixel = updatePixelRgba(&stream, pixel);
                                        *writer++ = (alpha << RIP_ASHIFT) | pixel;
                                }
                        }

                        if (x >= this->clipwidth)  break;
                        if (stream.get()) {
                                alpha = stream.get(7) * 2 + 1;
                        } else {
                                alpha = stream.get() ? 0xff : 0;
                        }
                } while (true);
                if (x > this->clipwidth) // for broken data
                        break;
        }
}


Rip::Rip()
{
        this->width = 0;
        this->height = 0;
        this->clipleft = 0;
        this->cliptop = 0;
        this->clipwidth = 0;
        this->clipheight = 0;
        this->kind = 0;
        this->size = 0;
        this->in = NULL;
}


Rip::~Rip()
{
        free(this->in);
}


/**
 * @return success
 */
bool
Rip::initialize(const uint8_t* buffer, size_t size)
{
        uint16_t v;
        int classNameLen;
        MemoryStream st(buffer, size);

        if (st.get32le() != 0x29f6cba4)  return false;
        if (st.isEof())  return false;
        v = st.get16le();
        if (st.isEof())  return false;
        if (v == 0x0010) {
                if (st.get16le() != 0xffff)
                        return false;
        } else if (v == 0x0011) {
                if (st.get16le() != 0x0001)
                        return false;
                if (st.get16le() != 0xffff)
                        return false;
        } else if (v != 0xffff)
                return false;
        if (st.get16le() != 0x0001)
                return false;
        if (st.isEof())  return false;
        classNameLen = st.get16le();
        if (st.isEof())  return false;
        for (int i = 0;  i < classNameLen; i++)
                st.get8();
        if (st.isEof())  return false;

        if (classNameLen == 4) { /* CRip */
                st.get32le();
                this->clipleft = st.get16le();
                this->cliptop = st.get16le();
                this->width = st.get16le();
                this->height = st.get16le();
                this->clipwidth = st.get16le();
                this->clipheight = st.get16le();
                this->kind = st.get32le();
                if ((this->kind & 0xff) == 2) {
                        this->clipwidth = this->width;
                        this->clipheight = this->height;
                        switch ((this->kind & 0x00ff0000) >> 16) {
                        case 1:
                                this->type = T_CRIP1;
                                return false; // not implemented
                        case 2:
                                this->type = T_CRIP2;
                                return false; // not implemented
                        case 3:
                                this->type = T_CRIP3;
                                break;
                        default:
                                return false;
                        }
                } else if ((this->kind & 0xff) == 3) {
                        this->type = T_CRIPA;
                } else
                        return false;
                this->size = st.get32le();
                st.get32le();
        } else if (classNameLen == 7) { /* CRip007 */
                st.get32le();
                this->width = st.get16le();
                this->height = st.get16le();
                this->clipleft = st.get16le();
                this->cliptop = st.get16le();
                this->clipwidth = st.get16le();
                this->clipheight = st.get16le();
                this->kind = st.get32le();
                if ((this->kind & 0xff) == 2) {
                        this->type = T_CRIP007;
                } else if ((this->kind & 0xff) == 3) {
                        this->type = T_CRIP007A;
                } else
                        return false;
                for (int i = 0; i < 4; i++)
                        this->compInfo[i] = st.get8();
                this->weightB = st.get8();
                this->weightG = st.get8();
                this->weightR = st.get8();
                this->size = st.get32le();
                st.get32le();
        } else
                return false;
        if (st.isEof())  return false;
        if (this->width == 0 || this->height == 0)
                return false;

        if (this->size == 0 || this->size == 0xffffffff)
                return false;

	this->in = (uint8_t*)malloc(this->size);
        if (this->in == NULL)
                return false;
        st.read(this->in, this->size);
        return true;
}


#ifndef USE_STATICTABLE
static bool tryLoading(const char* path)
{
        HMODULE hmMfc;
        HMODULE hmGmfc;
        HMODULE hmUnivui;
        char fullPath[MAX_PATH];
        FARPROC addr;

        if (strlen(path) + 11 > MAX_PATH)
                return false;

        strcpy(fullPath, path);
        strcat(fullPath, "Mfc42.dll");
        hmMfc = LoadLibrary(fullPath);
        if (hmMfc == NULL)
                return false;

        strcpy(fullPath, path);
        strcat(fullPath, "GMfc.dll");
        hmGmfc = LoadLibrary(fullPath);
        if (hmGmfc == NULL) {
                FreeLibrary(hmMfc);
                return false;
        }

        strcpy(fullPath, path);
        strcat(fullPath, "UnivUI.dll");
        hmUnivui = LoadLibrary(fullPath);
        if (hmUnivui == NULL) {
                FreeLibrary(hmGmfc);
                FreeLibrary(hmMfc);
                return false;
        }

        addr = GetProcAddress(hmUnivui, "?_GLOBAL_tblQuantTransfer@@3PAUCQuantTbl@@A");
        if (addr == NULL) {
                FreeLibrary(hmUnivui);
                FreeLibrary(hmGmfc);
                FreeLibrary(hmMfc);
                return false;
        }

        memcpy(colorTranslationTable, (const uint8_t*)addr, 256 * 6);

        FreeLibrary(hmUnivui);
        FreeLibrary(hmGmfc);
        FreeLibrary(hmMfc);
        return true;
}

static bool tryLoadingRegistry(const char* key)
{
        HKEY hkey = NULL;
        HRESULT result;
        char path[MAX_PATH];
        DWORD len = MAX_PATH;

        result = RegOpenKeyEx(HKEY_CURRENT_USER, key, 0, KEY_READ, &hkey);
        if (result != NO_ERROR)
                return false;
        result = RegQueryValueEx(hkey, "strAbsolutePath", NULL, NULL, (BYTE*)path, &len);
        if (result != NO_ERROR) {
                RegCloseKey(hkey);
                return false;
        }
        RegCloseKey(hkey);

        return tryLoading(path);
}

static const char* pathKeys[] = {
        "Software\\relic UGP Applications\\age\\N]މi\\Installation",
        "Software\\relic UGP Applications\\age\\NG߃t{CX\\Installation",
        "Software\\relic UGP Applications\\age\\AJl}jAbNXQb\\Installation",
        "Software\\relic UGP Applications\\age\\AJl}jAbNXRb\\Installation",
        "Software\\relic UGP Applications\\age\\AJl}jAbNXSb\\Installation",
        "Software\\relic UGP Applications\\age\\AJl}jAbNXTb\\Installation",
        "Software\\relic UGP Applications\\ruf\\LQ\\Installation",
	// N]iłÔ̂͐Fe[uႤۂ̂ŁADx
        // Vi͂
        "Software\\relic UGP Applications\\age\\N]މǐ\\Installation",
        "Software\\relic UGP Applications\\age\\AJl}jAbNXPb\\Installation",
        NULL
};
#endif

static void
loadTable()
{
#ifndef USE_STATICTABLE
        if (isTableInitialized)
                return;
        isTableInitialized = true;
        if (tryLoading(""))
                return;
        for (int i = 0; pathKeys[i] != NULL; i++) {
                if (tryLoadingRegistry(pathKeys[i]))
                        return;
        }
        MessageBox(NULL, "UnivUI.dll ̃[hɎs܂B\n摜ɃfR[hł܂B",
                   "Rip decoder", MB_OK | MB_ICONERROR);
#endif
}


void
Rip::decode(RipPixel *dst, size_t bytes_per_line) const
{
        switch (this->type) {
        case T_CRIP007:
                loadTable();
                this->decode007(dst, bytes_per_line, 0, 0);
                break;
        case T_CRIP007A:
                loadTable();
                this->decode007A(dst, bytes_per_line, 0, 0);
                break;
        case T_CRIP1:
        case T_CRIP2:
                break;          // Not implemented
        case T_CRIP3:
                this->decodeRgb3(dst, bytes_per_line);
                break;
        case T_CRIPA:
                this->decodeRgba(dst, bytes_per_line);
                break;
        }
}


size_t
Rip::getWidth() const
{
        return this->clipwidth;
}


size_t
Rip::getHeight() const
{
        return this->clipheight;
}


#ifdef _WIN32
void
Rip::getInfo(SusiePictureInfo* pic) const
{
        pic->left = this->clipleft;
        pic->top = this->cliptop;
        pic->width = this->clipwidth;
        pic->height = this->clipheight;
        pic->x_density = 1;
        pic->y_density = 1;
        pic->colorDepth = this->isAlpha() ? 32 : 24;
#if 1
        HLOCAL hl = LocalAlloc(LMEM_FIXED, 32);
        char* buf = (char*)LocalLock(hl);
        switch (this->type) {
        case T_CRIP1:
                strcpy(buf, "CRip/RGB-1");
                break;
        case T_CRIP2:
                strcpy(buf, "CRip/RGB-2");
                break;
        case T_CRIP3:
                strcpy(buf, "CRip/RGB-3");
                break;
        case T_CRIPA:
                strcpy(buf, "CRip/ARGB");
                break;
        case T_CRIP007:
                strcpy(buf, "CRip007/RGB");
                break;
        case T_CRIP007A:
                strcpy(buf, "CRip007/ARGB");
                break;
        }
        LocalUnlock(hl);
        pic->hInfo = hl;
#else
        pic->hInfo = NULL;
#endif
}

void
Rip::getInfo(BITMAPINFOHEADER* bmih) const
{
        bmih->biWidth = this->clipwidth;
        bmih->biHeight = this->clipheight;
        bmih->biPlanes = 1;
        bmih->biBitCount = 32;
        bmih->biCompression = BI_RGB;
        bmih->biSizeImage = 0;  // 0 for BI_RGB
        bmih->biXPelsPerMeter = 96 * 10000 / 254; // 96dpi?
        bmih->biYPelsPerMeter = 96 * 10000 / 254; // 96dpi?
        bmih->biClrUsed = 0;
        bmih->biClrImportant = 0;
}
#endif


bool
Rip::isSupported(const char* filename, const uint8_t* buffer, size_t size)
{
        if (size < 0x23)
                return false;
        if (memcmp(buffer, "\xa4\xcb\xf6\x29\x10\x00\xff\xff"
                   "\x01\x00\x07\x00\x43\x52\x69\x70"
                   "\x30\x30\x37", 19) == 0) {
                return buffer[0x23] == 0x02 || buffer[0x23] == 0x03;
        }
        if (memcmp(buffer, "\xa4\xcb\xf6\x29\x11\x00\x01\x00\xff\xff"
                   "\x01\x00\x07\x00\x43\x52\x69\x70"
                   "\x30\x30\x37", 21) == 0) {
                return buffer[0x25] == 0x02 || buffer[0x25] == 0x03;
        }
        if (memcmp(buffer, "\xa4\xcb\xf6\x29\x11\x00\x01\x00\xff\xff"
                   "\x01\x00\x04\x00\x43\x52\x69\x70", 18) == 0) {
                if (buffer[0x22] == 0x02 && buffer[0x24] == 0x03)
                        return true;
#if 0
                if (buffer[0x22] == 0x02 && buffer[0x24] == 0x02)
                        return true;
                if (buffer[0x22] == 0x02 && buffer[0x24] == 0x01)
                        return true;
#endif
                if (buffer[0x22] == 0x03)
                        return true;
        }
        if (memcmp(buffer, "\xa4\xcb\xf6\x29\x10\x00\xff\xff"
                   "\x01\x00\x04\x00\x43\x52\x69\x70", 16) == 0) {
                if (buffer[0x20] == 0x02 && buffer[0x22] == 0x03)
                        return true;
#if 0
                if (buffer[0x20] == 0x02 && buffer[0x22] == 0x02)
                        return true;
                if (buffer[0x20] == 0x02 && buffer[0x22] == 0x01)
                        return true;
#endif
                if (buffer[0x20] == 0x03)
                        return true;
        }
        if (memcmp(buffer, "\xa4\xcb\xf6\x29\xff\xff"
                   "\x01\x00\x04\x00\x43\x52\x69\x70", 14) == 0) {
                if (buffer[0x1e] == 0x02 && buffer[0x20] == 0x03)
                        return true;
#if 0
                if (buffer[0x1e] == 0x02 && buffer[0x20] == 0x02)
                        return true;
                if (buffer[0x1e] == 0x02 && buffer[0x20] == 0x01)
                        return true;
#endif
                if (buffer[0x1e] == 0x03)
                        return true;
        }
        return false;
}


#if 0
int
main(int ac, char *av[])
{
	Rip rip;
	int x, y;
	Pixel *p;
        FILE* fp = fopen(av[1], "rb");
        FILE* fpout = fopen(av[2], "wb");

        if (fp == NULL) {
            perror(av[1]);
            return 1;
        }
        if (fpout == NULL) {
            perror(av[2]);
            return 1;
        }

        //
        // load
        //
        FileStream fs(fp);
        if (!rip.initialize(&fs)) {
                fprintf(stderr, "Unknown format\n");
                return 1;
        }
        rip.load(&fs) {
                fprintf(stderr, "Loading failed\n");
                return 1;
        }

        //
        // decode
        //
	p = calloc(rip.getWidth() * rip.getHeight(), sizeof(Pixel));
	fprintf(stderr, "Start uncompress\n");
	rip.uncompress(p, rip.getWdith() * sizeof(Pixel));

        //
        // output
        //
	fprintf(fpout, "P6\n%d %d\n255\n", rip.getWidth(), rip.getHeight());
	for (y=0; y<rip.getWidth(); y++) {
		for (x=0; x<rip.getHeight(); x++) {
			fputc((*p & 0x00ff0000)>>16, fpout);
			fputc((*p & 0x0000ff00)>>8, fpout);
			fputc((*p & 0x000000ff), fpout);
			p++;
		}
	}
        fclose(fpout);

	return 0;
}
#endif
