/*	$Id: Rip.h,v 1.10 2001/10/07 17:34:22 kent Exp $	-*- c++ -*- */
/*
 * Copyright 2001 TAMURA Kent
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef __RIP_H__
#define __RIP_H__

#include "ImageDecoder.h"

class Rip: public ImageDecoder {
        enum { T_CRIP1, T_CRIP2, T_CRIP3,
               T_CRIPA, T_CRIP007, T_CRIP007A } type;
        uint16_t width, height; // Size of destination
        uint16_t clipleft, cliptop; // Expected location in the destination
        uint16_t clipwidth, clipheight; // Size of this image
        uint32_t kind;
        uint8_t compInfo[4];    // for CRip007
#define CI_TABLE	0
#define CI_VALCLIP	3
        uint8_t weightB;        // for CRip007
        uint8_t weightG;        // for CRip007
        uint8_t weightR;        // for CRip007
        uint32_t size;          // The size of compressed data
        uint8_t* in;            // Compressed data

        void decode007(uint32_t* dst, size_t bytes_per_line, int dstx, int dsty) const;
        void decode007A(uint32_t* dst, size_t bytes_per_line, int dstx, int dsty) const;
        void decodeRgb3(uint32_t* dst, size_t bytes_per_line) const;
        void decodeRgba(uint32_t* dst, size_t bytes_per_line) const;
public:
        Rip();
        virtual bool initialize(const uint8_t* buffer, size_t size);
        virtual ~Rip();
        virtual size_t getWidth() const;
        virtual size_t getHeight() const;
        virtual void decode(uint32_t* dst, size_t bytes_per_line) const;
        virtual bool isAlpha() const {
                return (this->kind & 0xff) == 3;
        }

#ifdef _WIN32
        virtual void getInfo(SusiePictureInfo* pic) const;
        virtual void getInfo(BITMAPINFOHEADER* bmih) const;
#endif
        static bool isSupported(const char* filename, const uint8_t* buffer, size_t size);
};

#endif // __RIP_H__


