/*	$Id: Rqd.cpp,v 1.2 2001/11/20 14:54:59 kent Exp $	-*- c++ -*- */
/*
 * Copyright 2001 TAMURA Kent
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "Rqd.h"
#include "StreamImpl.h"

bool
Rqd::isSupported(const char* filename, const uint8_t* buffer, size_t size)
{
        if (memcmp(buffer, "\xa4\xcb\xf6\x29\xff\xff\x01\x00\x04\x00"
                   "\x43\x52\x71\x64", 14) == 0)
                return true;
        if (memcmp(buffer, "\xa4\xcb\xf6\x29\x11\x00\x01\x00\xff\xff\x01\x00\x04\x00"
                   "\x43\x52\x71\x64", 18) == 0)
                return true;
        return false;
}


Rqd::Rqd()
{
        this->size = 0;
        this->in = NULL;
}


Rqd::~Rqd()
{
        free(this->in);
}


bool
Rqd::initialize(const uint8_t* buffer, size_t size)
{
        uint16_t v;

        if (!Rqd::isSupported("", buffer, size))
                return false;

        MemoryStream st(buffer, size);
        if (st.get32le() != 0x29f6cba4)
                return false;
        v = st.get16le();           // 0xffff or 0x0011 0x0001 0xffff
        if (v == 0x0011 && st.get16le() == 0x0001) {
                v = st.get16le();
        }
        if (v != 0xffff)
                return false;
        if (st.get16le() != 0x0001)
                return false;
        if (st.get16le() != 4)
                return false;
        st.get32le();           // CRqd
        st.get32le();           // ? for example: 0x00000048
        this->width = st.get16le();
        this->height = st.get16le();
        st.get32le();           // ? for example: 0x00000180
        st.get32le();           // ? 0x00000000 clipleft + clipright?
        st.get32le();           // clipwidth + clipheight?
        this->size = st.get32le();
        if (st.isEof())  return false;
        if (this->size == 0)  return false;

        this->in = reinterpret_cast<uint8_t*>(malloc(this->size));
        if (this->in == NULL)
                return false;
        st.read(this->in, this->size);
        return true;
}


size_t
Rqd::getWidth() const
{
        return this->width;
}


size_t
Rqd::getHeight() const
{
        return this->height;
}


void
Rqd::decode(uint32_t* dst, size_t bytes_per_line) const
{
        const uint32_t* rp = reinterpret_cast<const uint32_t*>(this->in);
        for (int i = 0;  i < this->height; i++) {
                size_t nelem = *rp;
                uint32_t* wp = dst + (this->height - i - 1) * bytes_per_line / sizeof(uint32_t);
                for (uint32_t j = 1; j < nelem; j++) {
                        uint32_t length = rp[j];
                        int len = length & 0xffff;
                        if ((length & 0xff000000) == 0) {
                                for (int k = 0;  k < len; k++)
                                        *wp++ = 0;
                        } else {
                                for (; len > 0; len--)
                                        *wp++ = rp[++j];
                        }
                }
                rp += nelem;
        }
}


#ifdef _WIN32
void
Rqd::getInfo(SusiePictureInfo* pic) const
{
        pic->left = 0;
        pic->top = 0;
        pic->width = this->width;
        pic->height = this->height;
        pic->x_density = 1;
        pic->y_density = 1;
        pic->colorDepth = 32;
        pic->hInfo = NULL;
}


void
Rqd::getInfo(BITMAPINFOHEADER* bmih) const
{
        bmih->biWidth = this->width;
        bmih->biHeight = this->height;
        bmih->biPlanes = 1;
        bmih->biBitCount = 32;
        bmih->biCompression = BI_RGB;
        bmih->biSizeImage = 0;  // 0 for BI_RGB
        bmih->biXPelsPerMeter = 96 * 10000 / 254; // 96dpi?
        bmih->biYPelsPerMeter = 96 * 10000 / 254; // 96dpi?
        bmih->biClrUsed = 0;
        bmih->biClrImportant = 0;
}
#endif // _WIN32
