/*	$Id: StreamImpl.h,v 1.6 2001/10/05 01:16:03 kent Exp $ -*- c++ -*-	*/
/*
 * Copyright 2001 TAMURA Kent
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef __STREAMIMPL_H__
#define __STREAMIMPL_H__

#include <stdio.h>
#include "Stream.h"


class FileStream: public Stream {
        FILE* fp;
        long offset;
public:
        /**
         * nꂽ FILE* ́ÃIuWFNg̃fXgN^ŕB
         */
        FileStream(FILE* fp, long offset);
        FileStream(FILE* fp);
        ~FileStream();
        virtual int get();
        virtual bool isEof();
        virtual void seek(long off, int whence);
};


class MemoryStream: public Stream {
        const uint8_t* buffer;
        size_t size;
        uint32_t offset;
public:
        MemoryStream(const uint8_t* buffer, size_t size);
        virtual int get();
        virtual bool isEof();
        virtual void seek(long off, int whence);
};


#ifdef _WIN32
#include <windows.h>
#include <winbase.h>

class Win32FileStream: public Stream {
#define W32FS_BUFFER_MAX 8192
        bool eof;
        HANDLE fd;
        long offset;
        uint8_t buffer[W32FS_BUFFER_MAX];
        size_t size;
        uint32_t reader;
        void fillBuffer();
public:
        /**
         * nꂽt@Cnh́ÃIuWFNg̃fXgN^ŕB
         */
        Win32FileStream(HANDLE fd, long offset);
        ~Win32FileStream();
        virtual int get();
        virtual bool isEof();
        virtual void seek(long off, int whence);
        virtual int read(uint8_t* buffer, size_t length);
};
#endif // _WIN32

#endif // __STREAMIMPL_H__
