/*	$Id: axrio.cpp,v 1.12 2001/11/22 16:39:11 kent Exp $	*/
/*
 * Copyright 2001 TAMURA Kent
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <new.h>
#include <stdio.h>
#include <stdlib.h>
#include <windows.h>
#include "config.h"
#include "RioArchive.h"
#include "StreamImpl.h"
#define ESPI_IMPLEMENTATION
#include "espi.h"
#include "version.h"

#if 1
                                // for debug
int message(const char* mes, int v) {
        char buf[257];
        sprintf(buf, "%s: %d (0x%x)", mes, v, v);
        MessageBox(NULL, buf, "axrio", MB_OK);
        return 0;
}
#endif


/**
 * Xg/Ȃ߂ϐ
 *
 * @see #loadConfiguration()
 * @see #storeConfiguration()
 * @see #dialogProc()
 * @see NameFilterImpl
 */
enum {SUFFIX_RIP = 0, SUFFIX_RQD, SUFFIX_S5I,
      SUFFIX_SIA, SUFFIX_AUDIO, SUFFIX_OTHER, SUFFIX_SIZE};
bool suffixMask[SUFFIX_SIZE] = {
        true, true, false, false, false, false
};


/**
 * suffixMask QƂt@CtB^
 *
 * @see #GetArchiveInfo()
 * @see #EspiGetArchiveInfo()
 */
class NameFilterImpl: public NameFilter {
        static bool endsWith(const char* name, int len, const char* suf) {
                int suffixLen = strlen(suf);
                if (len < suffixLen)
                        return false;
                return stricmp(name + len - suffixLen, suf) == 0;
        }

public:
        bool accept(const char* name) const {
                int len = strlen(name);
                if (endsWith(name, len, ".rip"))
                        return suffixMask[SUFFIX_RIP];
                if (endsWith(name, len, ".rqd"))
                        return suffixMask[SUFFIX_RQD];
                if (endsWith(name, len, ".s5i"))
                        return suffixMask[SUFFIX_S5I];
                if (endsWith(name, len, ".sia"))
                        return suffixMask[SUFFIX_SIA];
                if (endsWith(name, len, ".mp3") || endsWith(name, len, ".rha"))
                        return suffixMask[SUFFIX_AUDIO];
                return suffixMask[SUFFIX_OTHER];
        }
};


EXTERN_C __declspec(dllexport) int __stdcall
GetPluginInfo(int type, CHAR* buf, int buflen)
{
        switch (type) {
        case 0:
                if (buflen > 0)
                        buf[0] = '0';
                else
                        return 0;
                if (buflen > 1)
                        buf[1] = '0';
                else
                        return 1;
                if (buflen > 2)
                        buf[2] = 'A';
                else
                        return 2;
                if (buflen > 3)
                        buf[3] = 'M';
                else
                        return 3;
                if (buflen > 4)
                        buf[4] = 0;
                return 4;

        case 1:
                lstrcpyn(buf, "rUGP rio archive extracter v" HMVERSION, buflen);
                return strlen(buf);

        case 2:
                lstrcpyn(buf, "*.rio;01Vc;30CxgG;20wi;10L", buflen);
                return strlen(buf);

        case 3:
                lstrcpyn(buf, "rUGP rio archive", buflen);
                return strlen(buf);
        default:
                return 0;
        }
}


EXTERN_C __declspec(dllexport) int __stdcall
IsSupported(LPSTR filename, DWORD misc)
{
        if (HIWORD(misc) == 0)
                return EspiIsSupportedFile(filename, reinterpret_cast<HANDLE>(misc));
        else
                return EspiIsSupportedBuffer(filename, reinterpret_cast<const BYTE*>(misc), 2048);
}


EXTERN_C __declspec(dllexport) int __stdcall
GetArchiveInfo(LPSTR buf, long len, unsigned int flags, HLOCAL* ph)
{
        HSARCHIVE arc;
        int ret;

        if (SPI_IS_SOURCE_MEMORY(flags))
                ret = EspiOpenArchiveFromBuffer(reinterpret_cast<const BYTE*>(buf), len, &arc);
        else
                ret = EspiOpenArchiveFromFile(buf, len, &arc);
        if (ret)
                return ret;

        RioArchive* rio = reinterpret_cast<RioArchive*>(arc);
        ObjectOcean* oo = static_cast<ObjectOcean*>(rio->getTopObject());
        NameFilterImpl filter;
        int num = oo->getTotalNum(&filter);
        *ph = LocalAlloc(LMEM_MOVEABLE | LMEM_ZEROINIT, (num+1) * sizeof(SusieFileInfo));
        if (*ph == NULL)
                ret = SPI_ERROR_NO_MEMORY;
        else {
                SusieFileInfo* sfi = reinterpret_cast<SusieFileInfo*>(LocalLock(*ph));
                oo->storeAll(&filter, sfi)->method[0] = 0;
                LocalUnlock(*ph);
        }

        EspiCloseArchive(arc);
        return ret;
}


EXTERN_C __declspec(dllexport) int __stdcall
GetFileInfo(LPSTR buf, long len, LPSTR filename, unsigned int flags, SusieFileInfo* pf)
{
        HSARCHIVE arc;
        int ret;

        if (SPI_IS_SOURCE_MEMORY(flags))
                ret = EspiOpenArchiveFromBuffer(reinterpret_cast<const BYTE*>(buf), len, &arc);
        else
                ret = EspiOpenArchiveFromFile(buf, len, &arc);
        if (ret)
                return ret;
        ret = EspiGetFileInfo(arc, filename, SPI_IS_IGNORE_CASE(flags), pf);
        EspiCloseArchive(arc);
        return ret;
}


static int
GetFileToBuffer(RioArchive* rio, const ObjectOceanNode* node, HLOCAL* phl,
                PROGRESSCALLBACK callback, long ldata)
{
        if (node == NULL)
                return SPI_ERROR_BROKEN_DATA;
        *phl = LocalAlloc(LMEM_MOVEABLE | LMEM_ZEROINIT, node->getSize());
        if (*phl == NULL)
                return SPI_ERROR_NO_MEMORY;

        if (callback && callback(0, node->getSize(), ldata)) {
                LocalFree(*phl);  *phl = NULL;
                return SPI_ERROR_CALLBACK_ABORT;
        }
        void* buf = LocalLock(*phl);
        rio->extract(node, buf);
        LocalUnlock(*phl);
        if (callback && callback(node->getSize(), node->getSize(), ldata)) {
                LocalFree(*phl);  *phl = NULL;
                return SPI_ERROR_CALLBACK_ABORT;
        }
        return SPI_ERROR_SUCCESS;
}

static int
GetFileToFile(RioArchive* rio, const ObjectOceanNode* node, const char* dirname,
              PROGRESSCALLBACK callback, long ldata) {
        HLOCAL h;
        int ret = GetFileToBuffer(rio, node, &h, callback, ldata);
        if (ret)
                return ret;
        char* fn = reinterpret_cast<char*>(alloca(strlen(dirname) + 1 +
                                                  strlen(node->getName()) + 1));
        sprintf(fn, "%s\\%s", dirname, node->getName());
        FILE* pf = fopen(fn, "wb");
        if (pf == NULL) {
                LocalFree(h);
                return SPI_ERROR_READ; // READ?
        }
        void* buf = LocalLock(h);
        if (node->getSize() > 0)
                fwrite(buf, node->getSize(), 1, pf);
        fclose(pf);
        LocalUnlock(h);
        LocalFree(h);
        return ret;
}

EXTERN_C __declspec(dllexport) int __stdcall
GetFile(LPSTR src, long len, LPSTR dest, unsigned int flags,
        PROGRESSCALLBACK callback, long context)
{
        if (SPI_IS_SOURCE_MEMORY(flags)) {
#if 0
                MessageBox(NULL, "Memory Input is not supported.",
                           "axrio.spi", MB_OK | MB_ICONERROR);
#endif
                return SPI_ERROR_NOT_IMPLEMENTED;
        }

        HANDLE fd = CreateFileA(src, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING,
                                FILE_ATTRIBUTE_NORMAL | FILE_FLAG_RANDOM_ACCESS, NULL);
        if (fd == INVALID_HANDLE_VALUE) {
                return SPI_ERROR_READ;
        }

        int ret;
        try {
                do {
                        RioArchive rio(new Win32FileStream(fd, 0));
                        ObjectOcean* oo = static_cast<ObjectOcean*>(rio.load());
                        if (oo == NULL) {
                                ret = SPI_ERROR_READ;
                                break;
                        }
                        const ObjectOceanNode* node = oo->searchFor(len);
                        if (node == NULL) {
                                ret = SPI_ERROR_BROKEN_DATA; // XXX
                                break;
                        }
                        if (SPI_IS_DEST_MEMORY(flags))
                                ret = GetFileToBuffer(&rio, node,
                                                      reinterpret_cast<HLOCAL*>(dest),
                                                      callback, context);
                        else
                                ret = GetFileToFile(&rio, node, dest, callback, context);
                } while (FALSE);
        } catch (bad_alloc) {
                CloseHandle(fd);
                ret = SPI_ERROR_NO_MEMORY;
        }
        return ret;
}


/* ================================================================
 * 	ESPI entries
 * ================================================================ */

EXTERN_C __declspec(dllexport) int __stdcall
EspiIsSupportedFile(const char* filename, HANDLE fd)
{
        const int BUFFERSIZE = 2048;
        BYTE buffer[BUFFERSIZE];
        memset(buffer, 0, BUFFERSIZE);
        DWORD len;
        int s = 0;

        while (s < BUFFERSIZE && ReadFile(fd, buffer+s, BUFFERSIZE-s, &len, NULL)) {
                s += len;
        }
        return EspiIsSupportedBuffer(filename, buffer, s);
}

EXTERN_C __declspec(dllexport) int __stdcall
EspiIsSupportedBuffer(const char* filename, const BYTE* buffer, size_t size)
{
        if (size < 8)
                return FALSE;
        if (memcmp(buffer, "\xcd\x32\x6e\x59\xff\xff", 6) == 0)
                return TRUE;
        if (memcmp(buffer, "\xcd\x32\x6e\x59\x10\x00\xff\xff", 8) == 0)
                return TRUE;
        if (memcmp(buffer, "\xcd\x32\x6e\x59\x11\x00\x01\x00\xff\xff", 10) == 0)
                return TRUE;
        return FALSE;
}


EXTERN_C __declspec(dllexport) int __stdcall
EspiOpenArchiveFromFile(const char* filename, long offset, HSARCHIVE* phsa)
{
        FileStream* st = NULL;
        RioArchive* rio = NULL;

        *phsa = NULL;
        try {
                FILE* fp = fopen(filename, "rb");
                if (fp == NULL)
                        return SPI_ERROR_READ;
                st = new FileStream(fp, offset);
                rio = new RioArchive(st);
                if (rio->load() == NULL)
                        return SPI_ERROR_BROKEN_DATA;
                *phsa = reinterpret_cast<HSARCHIVE>(rio);
        } catch (bad_alloc) {
                if (rio != NULL)
                        delete rio;
                else
                        delete st;
                return SPI_ERROR_NO_MEMORY;
        }
        return SPI_ERROR_SUCCESS;
}


EXTERN_C __declspec(dllexport) int __stdcall
EspiOpenArchiveFromBuffer(const BYTE* buffer, size_t size, HSARCHIVE* phsa)
{
        MemoryStream* st = NULL;
        RioArchive* rio = NULL;

        *phsa = NULL;
        try {
                st = new MemoryStream(buffer, size);
                rio = new RioArchive(st);
                if (rio->load() == NULL)
                        return SPI_ERROR_BROKEN_DATA;
                *phsa = reinterpret_cast<HSARCHIVE>(rio);
        } catch (bad_alloc) {
                if (rio != NULL)
                        delete rio;
                else
                        delete st;
                return SPI_ERROR_NO_MEMORY;
        }
        return SPI_ERROR_SUCCESS;
}


EXTERN_C __declspec(dllexport) int __stdcall
EspiCloseArchive(HSARCHIVE arc)
{
        if (arc == NULL)
                return SPI_ERROR_SUCCESS;
        RioArchive* rio = reinterpret_cast<RioArchive*>(arc);
        delete rio;
        return SPI_ERROR_SUCCESS;
}


EXTERN_C __declspec(dllexport) int __stdcall
EspiGetArchiveInfo(HSARCHIVE arc, HLOCAL* phl)
{
        if (arc == NULL)
                return SPI_ERROR_INTERNAL;
        RioArchive* rio = reinterpret_cast<RioArchive*>(arc);
        ObjectOcean* oo = static_cast<ObjectOcean*>(rio->getTopObject());

        NameFilterImpl filter;
        int num = oo->getTotalNum(&filter);
        *phl = LocalAlloc(LMEM_MOVEABLE | LMEM_ZEROINIT, num * sizeof(SusieFileInfo));
        if (*phl == NULL)
                return SPI_ERROR_NO_MEMORY;
        SusieFileInfo* sfi = reinterpret_cast<SusieFileInfo*>(LocalLock(*phl));
        oo->storeAll(&filter, sfi);
        LocalUnlock(*phl);
        return SPI_ERROR_SUCCESS;
}


EXTERN_C __declspec(dllexport) int __stdcall
EspiGetFileInfo(HSARCHIVE arc, const char* filename, BOOL ignoreCase, SusieFileInfo* psfi)
{
        if (arc == NULL)
                return SPI_ERROR_INTERNAL;
        RioArchive* rio = reinterpret_cast<RioArchive*>(arc);
        ObjectOcean* oo = static_cast<ObjectOcean*>(rio->getTopObject());

        const ObjectOceanNode* node = oo->searchFor(filename, ignoreCase);
        if (node == NULL) {
                return SPI_ERROR_BROKEN_DATA;
        }
        node->store(NULL, psfi);
        return SPI_ERROR_SUCCESS;
}


EXTERN_C __declspec(dllexport) int __stdcall
EspiGetFileToFile(HSARCHIVE arc, const char* filename, BOOL ignoreCase, const char* destdir,
                  PROGRESSCALLBACK callback, long ldata)
{
        if (arc == NULL)
                return SPI_ERROR_INTERNAL;
        RioArchive* rio = reinterpret_cast<RioArchive*>(arc);
        ObjectOcean* oo = static_cast<ObjectOcean*>(rio->getTopObject());
        const ObjectOceanNode* node = oo->searchFor(filename, ignoreCase);
        if (node == NULL)
                return SPI_ERROR_BROKEN_DATA;
        return GetFileToFile(rio, node, destdir, callback, ldata);
}


EXTERN_C __declspec(dllexport) int __stdcall
EspiGetFileToBuffer(HSARCHIVE arc, const char* filename, BOOL ignoreCase, HLOCAL* phl,
                    PROGRESSCALLBACK callback, long ldata)
{
        if (arc == NULL)
                return SPI_ERROR_INTERNAL;
        RioArchive* rio = reinterpret_cast<RioArchive*>(arc);
        ObjectOcean* oo = static_cast<ObjectOcean*>(rio->getTopObject());
        const ObjectOceanNode* node = oo->searchFor(filename, ignoreCase);
        if (node == NULL)
                return SPI_ERROR_BROKEN_DATA;
        return GetFileToBuffer(rio, node, phl, callback, ldata);
}


/* ================================================================
 * 	ݒ_CAOƐݒ֘A
 * ================================================================ */

// _CAO\[X邽߂ɁACX^XnhۑĂȂƂȂB
// MinGW ƃ^CƂ̌ˍŃCCƗ]vȂƂKvB
HINSTANCE hInstance;
static void loadConfiguration();
static void storeConfiguration();
#ifdef __MINGW32__
EXTERN_C void __mingwthr_run_key_dtors (DWORD);
extern CRITICAL_SECTION __mingwthr_cs;
#endif

/**
 * DLL ̃Gg[|Cg
 */
EXTERN_C BOOL APIENTRY
DllMain (HINSTANCE hDllHandle, DWORD reason, LPVOID reserved)
{
#ifdef __MINGW32__
        switch (reason) {
        case DLL_PROCESS_ATTACH:
                InitializeCriticalSection(&__mingwthr_cs);
                break;
        case DLL_PROCESS_DETACH:
                __mingwthr_run_key_dtors(GetCurrentThreadId());
                DeleteCriticalSection(&__mingwthr_cs);
                break;
        case DLL_THREAD_ATTACH:
                break;
        case DLL_THREAD_DETACH:
                __mingwthr_run_key_dtors(GetCurrentThreadId());
                break;
        }
#endif
        switch (reason) {
        case DLL_PROCESS_ATTACH:
                hInstance = hDllHandle;
                loadConfiguration();
                break;
        }
        return TRUE;
}


#include "resource.h"

static const int dialogIds[SUFFIX_SIZE] = {
        IDC_CHECK_RIP, IDC_CHECK_RQD, IDC_CHECK_S5I,
        IDC_CHECK_SIA, IDC_CHECK_AUDIO, IDC_CHECK_OTHER
};
                                // ݒێ郌WXgL[
static const char* KEY_AXRIO = "Software\\hauN.org\\axrio";
                                // ƒl
static const char* suffixValueNames[SUFFIX_SIZE] = {
        "acceptRip", "acceptRqd", "acceptS5i",
        "acceptSia", "acceptAudio", "acceptOther"
};

static BOOL CALLBACK
dialogProc(HWND dialog, UINT message, WPARAM wparam, LPARAM lparam)
{
        switch (message) {
        case WM_INITDIALOG:
        {
                // Initialize dialog position
                RECT parentRect, myRect;
                HWND parent = GetParent(dialog);
                if (parent == NULL)
                        parent = GetDesktopWindow();
                GetWindowRect(parent, &parentRect);
                GetWindowRect(dialog, &myRect);
                int myWidth = myRect.right - myRect.left;
                int myHeight = myRect.bottom - myRect.top;
                int parentWidth = parentRect.right - parentRect.left;
                int parentHeight = parentRect.bottom - parentRect.top;
                SetWindowPos(dialog, HWND_TOP, parentRect.left + (parentWidth - myWidth)/2,
                             parentRect.top + (parentHeight - myHeight)/2, 0, 0, SWP_NOSIZE);

                // Initialize control state
                for (int i = 0;  i < SUFFIX_SIZE; i++) {
                        CheckDlgButton(dialog, dialogIds[i],
                                       suffixMask[i] ? BST_CHECKED : BST_UNCHECKED);
                }
                break;
        }

        case WM_COMMAND:
                switch (LOWORD(wparam)) {
                case IDOK:
                {
                        for (int i = 0; i < SUFFIX_SIZE; i++) {
                                suffixMask[i] = IsDlgButtonChecked(dialog, dialogIds[i]) == BST_CHECKED;
                        }
                        EndDialog(dialog, wparam);
                        return TRUE;
                }
                case IDCANCEL:
                        EndDialog(dialog, wparam);
                        return TRUE;
                }
                break;
        }
        return FALSE;
}

static void
loadConfiguration()
{
        HKEY hkey;
        LONG result;

        result = RegOpenKeyEx(HKEY_CURRENT_USER, KEY_AXRIO, 0, KEY_READ, &hkey);
        if (result != NO_ERROR)
                return;

        for (int i = 0; i < SUFFIX_SIZE; i++) {
                DWORD v;
                DWORD len = sizeof(v);
                result = RegQueryValueEx(hkey, suffixValueNames[i],
                                         NULL, NULL, reinterpret_cast<BYTE*>(&v), &len);
                if (result == NO_ERROR)
                        suffixMask[i] = v != 0;
        }
        RegCloseKey(hkey);
}

static void
storeConfiguration()
{
        HKEY hkey;
        LONG result;

        result = RegCreateKey(HKEY_CURRENT_USER, KEY_AXRIO, &hkey);
        if (result != NO_ERROR) {
                MessageBox(NULL, "WXg쐬s", "axrio", MB_OK | MB_ICONERROR);
                return;
        }
        for (int i = 0; i < SUFFIX_SIZE; i++) {
                DWORD v = suffixMask[i];
                RegSetValueEx(hkey, suffixValueNames[i], 0, REG_DWORD,
                              reinterpret_cast<const BYTE*>(&v), sizeof(v));
        }
        RegCloseKey(hkey);
}

EXTERN_C __declspec(dllexport) int __stdcall
ConfigurationDlg(HWND parent, int fnc)
{
        if (fnc != SPI_CONF_CONF)
                return SPI_ERROR_NOT_IMPLEMENTED;
        int ret = DialogBox(hInstance, MAKEINTRESOURCE(IDD_DIALOG),
                            parent, reinterpret_cast<DLGPROC>(dialogProc));
        if (ret == IDOK)
                storeConfiguration();
        return SPI_ERROR_SUCCESS;
}

