
/*	$Id: hmbench.cpp,v 1.8 2001/10/04 13:02:47 kent Exp $	*/
/*
 * Copyright 2001 TAMURA Kent
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <windows.h>
#include <stdio.h>
#include <process.h>
#include <stdint.h>
#include "espi.h"
#include "mmx.h"
#include "version.h"

#define TITLE "hmbench " HMVERSION

static const int NFRAMES = 16;
static const int CLIENT_WIDTH = 640;
static const int CLIENT_HEIGHT = 584;
static const int FRAME_WIDTH = 558;
static const int FRAME_HEIGHT = 584;
static HBITMAP frames[NFRAMES];
static uint32_t* frameBits[NFRAMES];
static const char* frameNames[NFRAMES] = {
    "LData\\ACD007pǗ\\rd\\k3A_0001.rip",
    "LData\\ACD007pǗ\\rd\\k3A_0002.rip",
    "LData\\ACD007pǗ\\rd\\k3A_0003.rip",
    "LData\\ACD007pǗ\\rd\\k3A_0004.rip",
    "LData\\ACD007pǗ\\rd\\k3A_0005.rip",
    "LData\\ACD007pǗ\\rd\\k3A_0006.rip",
    "LData\\ACD007pǗ\\rd\\k3A_0007.rip",
    "LData\\ACD007pǗ\\rd\\k3A_0008.rip",
    "LData\\ACD007pǗ\\rd\\k3A_0009.rip",
    "LData\\ACD007pǗ\\rd\\k3A_0010.rip",
    "LData\\ACD007pǗ\\rd\\k3A_0011.rip",
    "LData\\ACD007pǗ\\rd\\k3A_0012.rip",
    "LData\\ACD007pǗ\\rd\\k3A_0013.rip",
    "LData\\ACD007pǗ\\rd\\k3A_0014.rip",
    "LData\\ACD007pǗ\\rd\\k3A_0015.rip",
    "LData\\ACD007pǗ\\rd\\k3A_0016.rip",
};
static CRITICAL_SECTION csection;
static HBITMAP offScreen;
static void* offScreenBits;
static HLOCAL hLogoBits;
static BOOL hasMMX;

#define LOGO_WIDTH	800
#define LOGO_HEIGHT	202
#define LOGO_RIP	"Data\\25wi\\f\\OpTitle.rip"
#define KEY_KIMINST	"Software\\relic UGP Applications\\age\\N]މi\\Installation"


HBITMAP loadFramePicture(HSARCHIVE arc, int i) {
    const char* name = frameNames[i];
    HLOCAL content;
    if (EspiGetFileToBuffer(arc, name, TRUE, &content, NULL, 0)) {
        MessageBox(NULL, name, TITLE, MB_OK);
        return NULL;
    }

    HSPICTURE hpic;
    HLOCAL hbinfo, hbm;
    void* pimage = LocalLock(content);
    if (EspiOpenPictureFromBuffer((BYTE*)pimage, LocalSize(content), &hpic)) {
        MessageBox(NULL, "摜J܂łB", TITLE, MB_OK);
        LocalFree(content);
        return NULL;
    }

    if (EspiGetPicture(hpic, ESPI_CAP_TOPDOWN | ESPI_CAP_32BPP, &hbinfo, &hbm, NULL, 0)) {
        MessageBox(NULL, "摜WJł܂łB", TITLE, MB_OK);
        EspiClosePicture(hpic);
        LocalFree(content);
        return NULL;
    }
    EspiClosePicture(hpic);
    LocalFree(content);

    BITMAPINFO* pbmi = (BITMAPINFO*)LocalLock(hbinfo);
    HBITMAP hb = CreateDIBSection(NULL, pbmi, DIB_RGB_COLORS, (void**)&frameBits[i], NULL, 0);
    if (hb == NULL) {
        MessageBox(NULL, "摜̏Ɏs܂(s?)B", TITLE, MB_OK);
        LocalFree(hbm);
        LocalFree(hbinfo);
        return NULL;
    }
    memcpy(frameBits[i], LocalLock(hbm), LocalSize(hbm));
    LocalFree(hbm);
    LocalFree(hbinfo);
    return hb;
}

HLOCAL loadLogoPicture(HSARCHIVE arc, const char* name) {
    HLOCAL content;
    if (EspiGetFileToBuffer(arc, name, TRUE, &content, NULL, 0)) {
        MessageBox(NULL, name, TITLE, MB_OK);
        return NULL;
    }

    HSPICTURE hpic;
    HLOCAL hbinfo, hbm;
    void* pimage = LocalLock(content);
    if (EspiOpenPictureFromBuffer((BYTE*)pimage, LocalSize(content), &hpic)) {
        MessageBox(NULL, "摜J܂łB", TITLE, MB_OK);
        LocalFree(content);
        return NULL;
    }

    if (EspiGetPicture(hpic, ESPI_CAP_ALPHA32, &hbinfo, &hbm, NULL, 0)) {
        MessageBox(NULL, "摜WJł܂łB", TITLE, MB_OK);
        EspiClosePicture(hpic);
        LocalFree(content);
        return NULL;
    }
    EspiClosePicture(hpic);
    LocalFree(content);
    LocalFree(hbinfo);
    return hbm;
}

BOOL initialize() {
    hasMMX = mmxIsEnabled();
    InitializeCriticalSection(&csection);

    BITMAPINFO bmi;
    memset(&bmi, 0, sizeof(BITMAPINFO));
    bmi.bmiHeader.biSize = sizeof(bmi.bmiHeader);
    bmi.bmiHeader.biWidth = CLIENT_WIDTH;
    bmi.bmiHeader.biHeight = CLIENT_HEIGHT;
    bmi.bmiHeader.biPlanes = 1;
    bmi.bmiHeader.biBitCount = 32;
    offScreen = CreateDIBSection(NULL, &bmi, DIB_RGB_COLORS, &offScreenBits, NULL, 0);
    if (offScreen == NULL) {
        MessageBox(NULL, "XN[obt@̊mێsB", TITLE, MB_OK);
        return FALSE;
    }
    
    HKEY hkey = NULL;
    HRESULT result = RegOpenKeyEx(HKEY_CURRENT_USER, KEY_KIMINST, 0, KEY_READ, &hkey);
    if (result != NO_ERROR) {
        MessageBox(NULL, "wN]މix̃CXg[񂪂܂B", TITLE, MB_OK);
        return FALSE;
    }

    char rioPath[MAX_PATH];
    DWORD len = MAX_PATH;
    result = RegQueryValueEx(hkey, "strObjectOcean", NULL, NULL, (BYTE*)rioPath, &len);
    if (result != NO_ERROR) {
        RegCloseKey(hkey);
        MessageBox(NULL, "wN]މix̃CXg[񂪂܂B", TITLE, MB_OK);
        return FALSE;
    }
    RegCloseKey(hkey);

    HSARCHIVE arc;
    if (EspiOpenArchiveFromFile(rioPath, 0, &arc)) {
        MessageBox(NULL, "N].rio J܂łB", TITLE, MB_OK);
        return FALSE;
    }

    hLogoBits = loadLogoPicture(arc, LOGO_RIP);
    if (hLogoBits == NULL) {
        EspiCloseArchive(arc);
        return FALSE;
    }
    BOOL ret = TRUE;
    for (int i = 0; i < NFRAMES; i++) {
        frames[i] = loadFramePicture(arc, i);
        if (frames[i] == NULL) {
            ret = FALSE;
            break;
        }
    }
    EspiCloseArchive(arc);

    return ret;
}

void finalize() {
    DeleteCriticalSection(&csection);
    for (int i = 0; i < NFRAMES; i++)
        DeleteObject(frames[i]);
    if (offScreen)
        DeleteObject(offScreen);
    if (hLogoBits)
        LocalFree(hLogoBits);
}


static int currentNumber = 0;   // \Ăt[ԍ 0 - 15
static int yure = 0;            // vJn\ x`݈̂Ӗ
static float yuresult = -1;
static BOOL usedMMX;
enum { BENCH_NONE = 0, BENCH_NORMAL, BENCH_NOMMX};
static int inBench = BENCH_NONE;
static DWORD startTick = 0;
static const unsigned int BENCH_TIME = 10*1000; // 10sec

void makeScreen() {
    char str[100];
    int len;
    if (yuresult < 0 && !inBench) {
        len = sprintf(str, "Xy[XL[ĂˁBRAQAPA͂B");
    } else if (inBench) {
        len = sprintf(str, "v: %d h", yure / NFRAMES);
    } else {
        const char* mmxstate = "MMXȂ";
        if (usedMMX)
            mmxstate = "MMXL";
        else if (hasMMX)
            mmxstate = "MMXgp";
        len = sprintf(str, ": %.2f h/10b (%s)", yuresult, mmxstate);
    }

    EnterCriticalSection(&csection);
    HDC destdc = CreateCompatibleDC(NULL);
    HDC srcdc = CreateCompatibleDC(NULL);
    HGDIOBJ dorig = SelectObject(destdc, offScreen);
    HGDIOBJ sorig = SelectObject(srcdc, frames[currentNumber]);

    /**
     *  BitBlt()  memcpy() ɂ肵ĂAĒxȂ
     */
    BitBlt(destdc, 0, 0, FRAME_WIDTH, FRAME_HEIGHT, srcdc, 0, 0, SRCCOPY);
                                // [̑ȂhĂ
    RECT edge = { FRAME_WIDTH, 0, CLIENT_WIDTH, CLIENT_HEIGHT};
    FillRect(destdc, &edge, (HBRUSH)GetStockObject(BLACK_BRUSH));

                                // x`}[NȂASƃAt@
    if (inBench != BENCH_NONE) {
        uint32_t* logoBits = (uint32_t*)LocalLock(hLogoBits);
        int dbpl = BMP_BYTE_WIDTH(CLIENT_WIDTH, 32)/sizeof(uint32_t);
        int lbpl = BMP_BYTE_WIDTH(LOGO_WIDTH, 32)/sizeof(uint32_t);

#if 0
        /*
         * S̊gƃAt@
         *
         * startTick ̌oߎԂɏ]
         * 	0          -> LOGO_WIDTH*1.1
         * 	BENCH_TIME -> LOGO_WIDTH
         * ƃjAɕωB
         */
        DWORD diffTick = GetTickCount() - startTick;
        int currentWidth = LOGO_WIDTH + LOGO_WIDTH * (BENCH_TIME - diffTick) * 10 / BENCH_TIME / 100;
        // g͉()
#else
        const int currentWidth = LOGO_WIDTH;
#endif
        int logox = CLIENT_WIDTH/2 - currentWidth/2;
        int logoy = CLIENT_HEIGHT/2 - (LOGO_HEIGHT*currentWidth/LOGO_WIDTH)/2;

        for (int y = logoy; y < logoy + LOGO_HEIGHT*currentWidth/LOGO_WIDTH; y++) {
            if (y < 0) continue;
            if (y >= CLIENT_HEIGHT) break;
            int ly = (y - logoy) * LOGO_HEIGHT / (LOGO_HEIGHT*currentWidth/LOGO_WIDTH);
            int x = logox < 0 ? 0 : logox;
            uint32_t* pdst = (uint32_t*)offScreenBits + (CLIENT_HEIGHT - y - 1)*dbpl + x;
            const uint32_t* psrc = (uint32_t*)logoBits + (LOGO_HEIGHT - ly - 1)*lbpl;

            if (inBench == BENCH_NORMAL && hasMMX) {
#define MMX1
                /**
                 * x
                 *   MMX1 > MMX2 > m[}
                 * ł邪A債ĂȂBover 1GHz  CPU  +1 hƂxB
                 * AK6-2 ȍ~EPentium II ȍ~ MMX Zjbg2邻Ȃ̂ŁA
                 * 2sNZɏZlƂƑł邩ȂB
                 */
                __asm__ volatile ("pxor %mm0,%mm0");
#ifdef MMX1
                /* MMX 1
                 *  dst = (src-dst)* + dst
                 *  t|Z̊֌WŃ̐x𗎂ƂĂ
                 */
                for (; x < logox + currentWidth; x++) {
                    if (x >= CLIENT_WIDTH) break;
                    register uint32_t pixel = psrc[(x - logox) * LOGO_WIDTH / currentWidth];
                    uint32_t alpha = (pixel & 0xff000000) >> 25;
                    if (alpha == 0x7f) {
                        *pdst++ = pixel;
                    } else if (alpha > 1) {
                        __asm__ volatile ("
			movd %0,%%mm1
			movd (%1),%%mm2
			punpcklbw %%mm0,%%mm1
			punpcklbw %%mm0,%%mm2
			psubw %%mm2,%%mm1

			movd %2,%%mm3
			punpcklwd %%mm3,%%mm3
			punpckldq %%mm3,%%mm3
			pmullw %%mm3,%%mm1
			psraw $7,%%mm1
			paddw %%mm1,%%mm2
			packuswb %%mm0,%%mm2
			movd %%mm2,(%1)" : : "r"(pixel), "r"(pdst), "r"(alpha) : "%eax");
                        pdst++;
                    } else      // alpha == 0
                        pdst++;
                }
                __asm__ volatile ("emms");
#elif defined(MMX2)
                /* MMX 2
                 *  dst = src* + dst*(1-)
                 */
                for (; x < logox + currentWidth; x++) {
                    if (x >= CLIENT_WIDTH) break;
                    register uint32_t pixel = psrc[(x - logox) * LOGO_WIDTH / currentWidth];
                    uint32_t alpha = (pixel & 0xff000000) >> 24;
                    if (alpha == 0xff) {
                        *pdst++ = pixel;
                    } else if (alpha > 0) {
                        __asm__ volatile ("
			movd %0,%%mm1
			movd (%1),%%mm2
			punpcklbw %%mm0,%%mm1
			punpcklbw %%mm0,%%mm2

			movd %2,%%mm3
			negl %2
			punpcklwd %%mm3,%%mm3
			addl $0x101,%2
			punpckldq %%mm3,%%mm3
			movd %2,%%mm4
			punpcklwd %%mm4,%%mm4
			punpckldq %%mm4,%%mm4

			pmullw %%mm3,%%mm1
			pmullw %%mm4,%%mm2
			paddw %%mm2,%%mm1
			psrlw $8,%%mm1
			packuswb %%mm0,%%mm1
			movd %%mm1,(%1)" : : "r"(pixel), "r"(pdst), "r"(alpha) : "%eax");
                        pdst++;
                    } else          // alpha == 0
                        pdst++;
                }
                __asm__ volatile ("emms");
#endif
            } else {
                /* MMX gp
                 * ɒxc
                 * dst ̃f[^͔j
                 */
                for (; x < logox + currentWidth; x++) {
                    if (x >= CLIENT_WIDTH) break;
                    register uint32_t pixel = psrc[(x - logox) * LOGO_WIDTH / currentWidth];
                    uint32_t alpha = (pixel & 0xff000000) >> 24;
                    if (alpha == 0xff) {
                        *pdst++ = pixel;
                    } else if (alpha > 0) {
                        register uint32_t opix1, opix2;
                        register uint32_t npix1, npix2;

                        opix1 = *pdst;
                        opix2 = (opix1 >> 8) & 0xff;	// 000000gg
                        opix1 &= 0xff00ff;		// 00rr00bb
                        opix1 *= 0x100 - alpha;		// rrrrbbbb
                        opix2 *= 0x100 - alpha;		// 0000gggg
                        npix1 = (pixel & 0xff00ff) * alpha;	// rrrrbbbb
                        npix2 = (pixel & 0x00ff00) >> 8;// 000000gg
                        npix2 *= alpha;			// 0000gggg
                        npix1 = (opix1 + npix1) >> 8;	// 00RRRRBB
                        npix2 = npix2 + opix2;
                        *pdst++ = (npix1 & 0xff00ff) + (npix2 & 0x00ff00);
                    } else          // alpha == 0
                        pdst++;
                }
            }
        }
        LocalUnlock(hLogoBits);
    }
    SIZE size;
    GetTextExtentPoint32(destdc, str, len, &size);
    TextOut(destdc, CLIENT_WIDTH-size.cx, CLIENT_HEIGHT-size.cy, str, len);

    SelectObject(destdc, dorig);
    DeleteDC(destdc);
    SelectObject(srcdc, sorig);
    DeleteDC(srcdc);
    LeaveCriticalSection(&csection);
}


void benchThread(void* param) {
    HWND hwnd = (HWND)param;

    startTick = GetTickCount();
    yure = 0;
    do {
        if (++currentNumber == NFRAMES) {
            currentNumber = 0;
        }
        makeScreen();
        InvalidateRect(hwnd, NULL, FALSE);
        UpdateWindow(hwnd);
        yure++;
    } while (GetTickCount() - startTick < BENCH_TIME);
    yure--;                     // 10bȓɕ\m肽̂ŁA
                                // 10bo߂̖1
    yuresult = (float)yure / NFRAMES;
    inBench = BENCH_NONE;
}


LRESULT CALLBACK wndProc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam) {
    const UINT ANIMTIMER = 1;

    switch (msg) {

      case WM_CREATE:
        makeScreen();
        UpdateWindow(hwnd);
        SetTimer(hwnd, ANIMTIMER, 1000 / NFRAMES, NULL);
        break;

      case WM_DESTROY:
        PostQuitMessage(0);
        break;

      case WM_PAINT:
        EnterCriticalSection(&csection);
        {
            PAINTSTRUCT paint;
            HDC dc = BeginPaint(hwnd, &paint);
            HDC bitmapdc = CreateCompatibleDC(dc);
            HGDIOBJ orig = SelectObject(bitmapdc, offScreen);

            BitBlt(dc, 0, 0, CLIENT_WIDTH, CLIENT_HEIGHT, bitmapdc, 0, 0, SRCCOPY);

            SelectObject(bitmapdc, orig);
            DeleteDC(bitmapdc);
            EndPaint(hwnd, &paint);
        }
        LeaveCriticalSection(&csection);
        break;

      case WM_CHAR:
        if (!inBench) {
            switch (wparam) {
              case 0x1b:        // ESC
                DestroyWindow(hwnd);
                break;
              case ' ':
                yure = 0;
                yuresult = -1;
                InvalidateRect(hwnd, NULL, FALSE);
                inBench = BENCH_NORMAL;
                usedMMX = hasMMX;
                _beginthread(benchThread, 0, hwnd);
                break;
              case 'n':
                yure = 0;
                yuresult = -1;
                InvalidateRect(hwnd, NULL, FALSE);
                inBench = BENCH_NOMMX;
                usedMMX = FALSE;
                _beginthread(benchThread, 0, hwnd);
                break;
            }
        }
        break;

      case WM_TIMER:
        if (wparam == ANIMTIMER) {
            if (!inBench) {
                if (++currentNumber == NFRAMES) {
                    currentNumber = 0;
                }
                makeScreen();
                InvalidateRect(hwnd, NULL, FALSE);
                UpdateWindow(hwnd);
            }
        }
        break;

      default:
        return DefWindowProc(hwnd, msg, wparam, lparam);
    }
    return 0;
}


int APIENTRY WinMain(HINSTANCE hinst, HINSTANCE, LPSTR commandLine, int cmdShow) {

    if (!initialize()) {
        return 0;
    }

    WNDCLASS wc;
    memset(&wc, 0, sizeof(WNDCLASS));
    wc.lpfnWndProc = wndProc;
    wc.hInstance = hinst;
    wc.hbrBackground = (HBRUSH)GetStockObject(BLACK_BRUSH);
    wc.lpszClassName = "hmBenchClass";
    if (!RegisterClass(&wc)) {
        finalize();
        MessageBox(NULL, "s(EBhENX̓o^)", TITLE, MB_OK | MB_ICONERROR);
        return 0;
    }

    RECT cl = {0, 0, CLIENT_WIDTH, CLIENT_HEIGHT};
    AdjustWindowRect(&cl, WS_CAPTION | WS_SYSMENU, FALSE);
    HWND hwnd = CreateWindow("hmBenchClass", TITLE, WS_CAPTION | WS_SYSMENU,
                             CW_USEDEFAULT, CW_USEDEFAULT, cl.right - cl.left, cl.bottom - cl.top,
                             NULL, NULL, hinst, 0);
    if (hwnd == NULL) {
        finalize();
        MessageBox(NULL, "s(EBhE̍쐬)", TITLE, MB_OK | MB_ICONERROR);
        return 0;
    }
    ShowWindow(hwnd, SW_SHOW);

    MSG mes;
    int ret;
    while ((ret = GetMessage(&mes, NULL, 0, 0))) {
        if (ret == -1) {
            //MessageBox(NULL, "bZ[W[vŃG[", "hmbench", MB_OK | MB_ICONERROR);
            break;
        }
        TranslateMessage(&mes);
        DispatchMessage(&mes);
    }

    finalize();
    return mes.wParam;
}
