/*	$Id: ifrip.cpp,v 1.17 2001/11/20 15:02:21 kent Exp $	*/
/*
 * Copyright 2001 TAMURA Kent
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <new.h>
#include <assert.h>
#include <sys/stat.h>
#include "version.h"
#define ESPI_IMPLEMENTATION
#include "espi.h"
#include "ImageDecoderFactory.h"
#include "StreamImpl.h"


#if 0
                                // for debug
static int message(const char* mes, int v) {
        char buf[257];
        sprintf(buf, "%s: %d (0x%x)", mes, v, v);
        MessageBox(NULL, buf, "ifrip", MB_OK);
        return 0;
}
#endif


EXTERN_C __declspec(dllexport) int __stdcall
GetPluginInfo(int type, CHAR* buf, int buflen)
{
        assert(sizeof(SusiePictureInfo) == 26);
        switch (type) {
        case 0:
                if (buflen > 0)
                        buf[0] = '0';
                else
                        return 0;
                if (buflen > 1)
                        buf[1] = '0';
                else
                        return 1;
                if (buflen > 2)
                        buf[2] = 'I';
                else
                        return 2;
                if (buflen > 3)
                        buf[3] = 'N';
                else
                        return 3;
                if (buflen > 4)
                        buf[4] = 0;
                return 4;

        case 1:
                lstrcpyn(buf, "rUGP CRip/CRip007/CRqd decoder v" HMVERSION, buflen);
                return strlen(buf);

        case 2:
                lstrcpyn(buf, "*.rip;*.rqd", buflen);
                return strlen(buf);

        case 3:
                lstrcpyn(buf, "rUGP CRip/CRqd image", buflen);
                return strlen(buf);
        default:
                return 0;
        }
}


EXTERN_C __declspec(dllexport) int __stdcall
IsSupported(LPSTR filename, DWORD misc)
{
        if (HIWORD(misc) == 0)
                return EspiIsSupportedFile(filename,
                                           reinterpret_cast<HANDLE>(misc));
        else
                return EspiIsSupportedBuffer(filename,
                                             reinterpret_cast<const BYTE*>(misc),
                                             2048);
}


EXTERN_C __declspec(dllexport) int __stdcall
GetPictureInfo(LPSTR buf, long len, unsigned int flags, SusiePictureInfo* pspi)
{
        HSPICTURE pic;
        int ret;

        if (SPI_IS_SOURCE_MEMORY(flags))
                ret = EspiOpenPictureFromBuffer(reinterpret_cast<const BYTE*>(buf),
                                                len, &pic);
        else
                ret = EspiOpenPictureFromFile(buf, len, &pic);
        if (ret) {
                return ret;
        }
        ret = EspiGetPictureInfo(pic, pspi);

        EspiClosePicture(pic);
        return ret;
}


EXTERN_C __declspec(dllexport) int __stdcall
GetPicture(LPSTR buf, long len, unsigned int flags, HANDLE* phbmi, HANDLE* phbitmap,
           PROGRESSCALLBACK callback, long ldata)
{
        HSPICTURE pic;
        int ret;

        if (SPI_IS_SOURCE_MEMORY(flags))
                ret = EspiOpenPictureFromBuffer(reinterpret_cast<const BYTE*>(buf),
                                                len, &pic);
        else
                ret = EspiOpenPictureFromFile(buf, len, &pic);
        if (ret) {
                return ret;
        }
        ret = EspiGetPicture(pic, ESPI_CAP_NONE, phbmi, phbitmap, callback, ldata);
        EspiClosePicture(pic);
        return ret;
}


EXTERN_C __declspec(dllexport) int __stdcall
GetPreview(LPSTR buf, long len, unsigned int flags, HANDLE* phbmi, HANDLE* phbitmap,
           PROGRESSCALLBACK callback, long ldata)
{
        //return GetPicture(buf, len, flags, phbmi, phbitmap, callback, ldata);
        return SPI_ERROR_NOT_IMPLEMENTED;
}


/* ================================================================
 * 	ESPI entries
 * ================================================================ */

EXTERN_C __declspec(dllexport) int __stdcall
EspiIsSupportedFile(const char* filename, HANDLE fd)
{
        const int BUFFERSIZE = 2048;
        BYTE buffer[BUFFERSIZE];
        memset(buffer, 0, BUFFERSIZE);
        DWORD len;
        int s = 0;

        while (s < BUFFERSIZE && ReadFile(fd, buffer+s, BUFFERSIZE-s, &len, NULL)) {
                s += len;
        }
        return EspiIsSupportedBuffer(filename, buffer, s);
}


EXTERN_C __declspec(dllexport) int __stdcall
EspiIsSupportedBuffer(const char* filename, const BYTE* buffer, size_t size)
{
        return ImageDecoderFactory::isSupported(filename, buffer, size);
}


EXTERN_C __declspec(dllexport) int __stdcall
EspiOpenPictureFromFile(const char* filename, long offset, HSPICTURE* phsp)
{
        struct stat st;
        FILE* fp = NULL;
        BYTE* buffer;
        size_t size;
        int ret;

        *phsp = NULL;
        if (stat(filename, &st))
                return SPI_ERROR_READ;
        size = st.st_size;
                
        if ((fp = fopen(filename, "rb")) == NULL)
                return SPI_ERROR_READ;
                
        try {
                buffer = new BYTE[size];
        } catch (bad_alloc) {
                fclose(fp);
                return SPI_ERROR_NO_MEMORY;
        }
        if (size != fread(buffer, sizeof(BYTE), size, fp)) {
                delete buffer;
                fclose(fp);
                return SPI_ERROR_READ;
        }
        ret = EspiOpenPictureFromBuffer(buffer, size, phsp);
        delete buffer;
        fclose(fp);
        return ret;
}


EXTERN_C __declspec(dllexport) int __stdcall
EspiOpenPictureFromBuffer(const BYTE* buffer, size_t size, HSPICTURE* phsp)
{
        *phsp = NULL;
        ImageDecoder* rip = ImageDecoderFactory::create(buffer, size);
        if(rip == NULL) {
                return SPI_ERROR_UNKNOWN_FORMAT;
        }
        *phsp = reinterpret_cast<HSPICTURE>(rip);
        return SPI_ERROR_SUCCESS;
}


EXTERN_C __declspec(dllexport) int __stdcall
EspiClosePicture(HSPICTURE pic)
{
        if (pic != NULL) {
                ImageDecoder* rip = reinterpret_cast<ImageDecoder*>(pic);
                delete rip;
        }
        return SPI_ERROR_SUCCESS;
}


EXTERN_C __declspec(dllexport) int __stdcall
EspiGetPictureInfo(HSPICTURE pic, SusiePictureInfo* pspi)
{
        if (pic == NULL)
                return SPI_ERROR_INTERNAL;
        ImageDecoder* rip = reinterpret_cast<ImageDecoder*>(pic);

        rip->getInfo(pspi);
        return SPI_ERROR_SUCCESS;
}


EXTERN_C __declspec(dllexport) int __stdcall
EspiGetPicture(HSPICTURE pic, DWORD cap, HLOCAL* phbmi, HLOCAL* phbitmap,
               PROGRESSCALLBACK callback, long ldata)
{
        if (pic == NULL)
                return SPI_ERROR_INTERNAL;
        ImageDecoder* rip = reinterpret_cast<ImageDecoder*>(pic);

        *phbmi = LocalAlloc(LMEM_MOVEABLE | LMEM_ZEROINIT,
                            sizeof(BITMAPINFO) + sizeof(RGBQUAD) * 3);
        if (*phbmi == NULL)
                return SPI_ERROR_NO_MEMORY;
        BITMAPINFO* bi = reinterpret_cast<BITMAPINFO*>(LocalLock(*phbmi));
        bi->bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
        rip->getInfo(&bi->bmiHeader);

        int bpl32 = BMP_BYTE_WIDTH(bi->bmiHeader.biWidth, 32);
        *phbitmap = LocalAlloc(LMEM_MOVEABLE | LMEM_ZEROINIT,
                               bpl32 * rip->getHeight());
        if (*phbitmap == NULL) {
                LocalUnlock(*phbmi);
                return SPI_ERROR_NO_MEMORY;
        }

        if (callback && callback(0, rip->getHeight(), ldata))
                return SPI_ERROR_CALLBACK_ABORT;

        BYTE* pbmp = reinterpret_cast<BYTE*>(LocalLock(*phbitmap));
        rip->decode(reinterpret_cast<uint32_t*>(pbmp), bpl32);

        bool to24 = false;
        if (rip->isAlpha()) {
                // CAP_32BPP -> as is
                // CAP_ALPHA32 -> as is
                // neither -> to 24bpp
                if (cap & ESPI_CAP_32BPP) {
                } else if (cap & ESPI_CAP_ALPHA32) {
                        bi->bmiHeader.biCompression = BI_BITFIELDS;
                        bi->bmiColors[0].rgbBlue = 0xff; // Zero-cleard.
                        bi->bmiColors[1].rgbGreen = 0xff;
                        bi->bmiColors[2].rgbRed = 0xff;
                        bi->bmiColors[3].rgbReserved = 0xff;
                } else {
                        to24 = true;
                }
        } else {
                // CAP_32BPP -> as is
                // CAP_ALPHA32 -> to 24bpp
                // neither -> to 24bpp
                if (cap & ESPI_CAP_32BPP) {
                } else {
                        to24 = true;
                }
        }

        if (to24) {
                if (callback && callback(rip->getHeight()/2, rip->getHeight(), ldata)) {
                        LocalFree(*phbmi);  *phbmi = NULL;
                        LocalFree(*phbitmap);  *phbitmap = NULL;
                        return SPI_ERROR_CALLBACK_ABORT;
                }
                // Fix 32bit dib to 24bit dib
                int bpl24 = BMP_BYTE_WIDTH(bi->bmiHeader.biWidth, 24);
                bi->bmiHeader.biBitCount = 24;
                for (UINT y = 0; y < rip->getHeight(); y++) {
                        const BYTE* src = pbmp + y * bpl32;
                        BYTE* dst = pbmp + y*bpl24;
                        for (int x = 0; x < bi->bmiHeader.biWidth; x++) {
                                *dst++ = *src++;
                                *dst++ = *src++;
                                *dst++ = *src++;
                                src++;
                        }
                }
                LocalUnlock(*phbmi);
                LocalUnlock(*phbitmap);
                LocalReAlloc(*phbitmap, bpl24 * rip->getHeight(), LMEM_MOVEABLE);
        } else {
                LocalUnlock(*phbmi);
                LocalUnlock(*phbitmap);
        }
        if (callback && callback(rip->getHeight(), rip->getHeight(), ldata)) {
                LocalFree(*phbmi);  *phbmi = NULL;
                LocalFree(*phbitmap);  *phbitmap = NULL;
                return SPI_ERROR_CALLBACK_ABORT;
        }
        return SPI_ERROR_SUCCESS;
}

EXTERN_C __declspec(dllexport) int __stdcall
EspiGetPreview(HSPICTURE pic, DWORD cap, HLOCAL* phbinfo, HLOCAL* phbm,
               PROGRESSCALLBACK callback, long ldata)
{
        return SPI_ERROR_NOT_IMPLEMENTED;
}

